/*
Sand3d
Copyright (C) 2007 - David Pascoe-Deslauriers

  This file is part of Sand3d.

    Sand3d is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Sand3d is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sand3d; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <iostream>
#include <cstdlib>
#include "world.h"
#include "loadTextures.h"

using namespace std;

// compatibility with original GLUT
#if !defined(GLUT_WHEEL_UP)
#  define GLUT_WHEEL_DOWN 4
#  define GLUT_WHEEL_UP   5
#endif

world* w;

GLuint	texture[4];

int windx, windy;

int point_x,point_y,point_z;
int selected_type;
bool sand_on;
bool pause;
int graphics;
bool camera_on;

int pen_size;

int mouse_x, mouse_y;
double polar_x, polar_y, polar_d;

double fps;
int frames;
double timeEl;
int particleCount;

#define NUM_F 5
fountain *f[NUM_F];

#define NUM_E 21
fountain *exp[NUM_E];


int a = 100;
int b = 50;
int c = 100;

float light_ambient[]  = { 0.0f, 0.0f, 0.0f, 1.0f };
float light_diffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
float light_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
float light_position[] = { -50.0f, 50.0f, -50.0f, 0.0f };
float mat_ambient[]    = { 0.7f, 0.7f, 0.7f, 1.0f };
float mat_diffuse[]    = { 0.8f, 0.8f, 0.8f, 1.0f };
float mat_specular[]   = { 1.0f, 1.0f, 1.0f, 1.0f };
float high_shininess[] = { 40.0f };

void cycle_type(int type,int num){
    if (num == -1) return;
    selected_type = (selected_type + 1) % w->ptypes;
    if (type == -1) return;
    if (type == w->types[selected_type].phase) return;
    cycle_type(type,num-1);
}    

void add_sand(){
    switch (pen_size){
        case 0:
            pen_size = 1;
        break;
        case 1:
            w->add_particle(selected_type,point_x,point_y,point_z);
            if (w->types[selected_type].phase != w->SKYHOOK) pen_size = 0;
        break;
        default:    
            if (w->types[selected_type].phase == w->GENERATOR || w->types[selected_type].phase == w->SKYHOOK || w->types[selected_type].phase == w->HOLLOW){
                for(int a = 0; a < pen_size; a++){
                    for (int b = 0; b < pen_size; b++){
                        for (int c = 0; c < pen_size; c++){
                            if (w->types[selected_type].phase == w->HOLLOW && a > 0 && a < pen_size-1 && b > 0 && b < pen_size-1 && c > 0 && c < pen_size-1);
                            else w->add_particle(selected_type,point_x+a-(pen_size-1)/2,point_y+b-(pen_size-1)/2,point_z+c-(pen_size-1)/2);
                        }
                    }
                }            
            } else {
                for (int i = 0; i < pen_size - 1;i++){
                    int xp,yp,zp;
                    xp = rand()%3-1;
                    yp = rand()%3-1;
                    zp = rand()%3-1;
                    if (xp == 0){
                        zp += rand()%3-1;
                    } else if (zp == 0){
                        xp += rand()%3-1;
                    }
                    w->add_particle(selected_type,point_x+xp,point_y+yp,point_z+zp);
                }
            }    
        break;
    }    
}

void render_text(const char* string,double a,double b,double c){
    glDisable(GL_LIGHTING);
	char* p;
	float width = 0;
   
   	p = (char*) string;
	glRasterPos3f(a,b,c);
  	while (*p != '\0')
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *p++);
    glEnable(GL_LIGHTING);
}

/* GLUT callback Handlers */

static void resize(int width, int height) {
    windx = width;
    windy = height;
    const float ar = (float) width / (float) height;
    
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-ar, ar, -1.0, 1.0, 2.0, 1000.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity() ;
}

static void display(void) {
    if (sand_on) add_sand();
    
    const double t = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
    frames++;
    if (frames%5==0){
        fps = 5 / (t - timeEl);
        timeEl = t;
    }    
       
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

    glColor4f(1,1,1,1);
    char sfps[48];
    char part[48];
    sprintf(sfps,"%f",fps);
    sprintf(part,"%d",particleCount);

    glPushMatrix();
        glTranslated(0.9,0.9,-2);
        render_text(sfps,0,0,0);
    glPopMatrix();
    glPushMatrix();
        glTranslated(0.9,0.8,-2);
        render_text(part,0,0,0);
    glPopMatrix();
      
    if (polar_d < 10) polar_d = 10;

    glTranslated(0,0,-1*polar_d);
      
    if (polar_x >= 360)
        polar_x -= 360;
    if (polar_x < 0)
        polar_x += 360;

    if (polar_y >= 360)
        polar_y -= 360;
    if (polar_y < 0)
        polar_y += 360;
    glRotated(polar_x,1,0,0);
    glRotated(polar_y,0,1,0);
    
    glTranslated(-1*w->x/2,0,-1*w->z/2); 
    light_position[0]=-1*w->x/2;
    light_position[1]= w->y*2;
    light_position[2]=-1*w->z/2;
    glLightfv(GL_LIGHT0, GL_AMBIENT,  light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);

    glMaterialfv(GL_FRONT, GL_AMBIENT,   mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE,   mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR,  mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess); 
   
    glPushMatrix();
          glTranslatef(0, 0, 0);
          if (texture[3]){
                    glColor4f(1, 1, 1,1);
                    glEnable(GL_TEXTURE_2D);
                    glBindTexture(GL_TEXTURE_2D,texture[3]);
          } else {
                    glColor4f(0.25, 0.25, 0.25,1);          
          }    
          
          glDisable(GL_LIGHTING);
          
          glBegin(GL_TRIANGLE_STRIP);
                    glTexCoord2d(1,0); glVertex3f(w->x-0.5, 0.0f, w->z-0.5);
                    glTexCoord2d(1,1); glVertex3f(w->x-0.5, 0.0f, -0.5f);
                    glTexCoord2d(0,0); glVertex3f(-0.5f, 0.0f, w->z-0.5);
                    glTexCoord2d(0,1); glVertex3f(-0.5f, 0.0f, -0.5f);
          glEnd();
          glDisable(GL_TEXTURE_2D);
          glEnable(GL_LIGHTING);
    glPopMatrix();
    
    glPushMatrix();
          glColor4f(1, 0, 0,1);
          glTranslated(point_x,point_y,point_z);
          int size = pen_size;
          if (size == 0) size = 1;
          for (int i = point_y; i >= 0; i=i-size){
              if (i == point_y){
                       glColor4f(1, 1, 1,1);
              } else {    
                    glColor4f(1, 0, 0,1);
                    glTranslated(0,-size,0);
              }    
                    glutWireCube(size);
          }
    glPopMatrix();
    
    particleCount = 0;

    for (int b = 0; b < w->y; b++){
        for (int a = 0; a < w->x; a++){
            for(int c = 0; c < w->z; c++){
                if (w->w[a][b][c].type != world::EMPTY){
                            particleCount++;
                            
                            glPushMatrix();
                                    glTranslated(a,b,c);
                                    if (!w->w[a][b][c].fire){
                                        if (!w->w[a][b][c].explodes){
                                            if (w->w[a][b][c].fireid < 0){
                                                w->w[a][b][c].fireid = rand()%NUM_F; 
                                            }
                                            glPushMatrix();    
                                                glRotated(w->w[a][b][c].yrot,0,1,0);
                                                f[w->w[a][b][c].fireid]->draw(texture[0]);
                                            glPopMatrix();
                                        } else {
                                            if (w->w[a][b][c].fireid == w->count_exp || w->w[a][b][c].fireid > NUM_E) {
                                                w->add_particle(world::EMPTY,a,b,c);
                                            } else {
                                                if (w->w[a][b][c].fireid < 0) {
                                                    w->w[a][b][c].fireid = w->count_exp; 
                                                    w->w[a][b][c].yrot = rand()%360;
                                                }
                                                glPushMatrix();    
                                                    glRotated(w->w[a][b][c].yrot,0,1,0);
                                                    exp[w->w[a][b][c].fireid]->draw(texture[0]);
                                                glPopMatrix();
                                            }
                                        }        
                                    } else {
                                        glColor4f(w->w[a][b][c].colour.x, w->w[a][b][c].colour.y, w->w[a][b][c].colour.z, w->w[a][b][c].alpha);
                                        double lsize = 1.25;

                                        switch(graphics){
                                            default:
                                            case 0:
                                                switch(w->w[a][b][c].phase){
                                                    default:
                                                    case world::SOLID:
                                                        glutSolidSphere(0.75,7,7);
                                                        break;
                                                    case world::HOLLOW:    
                                                    case world::SKYHOOK:
                                                    case world::GENERATOR:
                                                        glutSolidCube(1);
                                                        break;
                                                    case world::LIQUID:
                                                        glDisable(GL_DEPTH_TEST);
                                                        glBlendFunc(GL_ONE_MINUS_DST_COLOR,GL_ONE);
                                                        glEnable(GL_TEXTURE_2D);

                                                        glBindTexture(GL_TEXTURE_2D,texture[1]);
                                            			glPushMatrix();
                                            			GLfloat matrix[16];
                                            			glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
                                            			matrix[0] = 1.0f; matrix[1] = 0.0f; matrix[2] = 0.0f;
                                            			matrix[4] = 0.0f; matrix[5] = 1.0f; matrix[6] = 0.0f;
                                            			glLoadMatrixf(matrix);
                                            			
                                            			glBindTexture(GL_TEXTURE_2D,texture[1]);
                                            			glBegin(GL_TRIANGLE_STRIP);
                                            			    glTexCoord2d(1,1); glVertex3f(lsize,lsize,0);
                                            				glTexCoord2d(0,1); glVertex3f(-1*lsize,lsize,0);
                                            				glTexCoord2d(1,0); glVertex3f(lsize,-1*lsize,0);
                                            				glTexCoord2d(0,0); glVertex3f(-1*lsize,-1*lsize,0);
                                            			glEnd();
  			
                                            			glRotated(rand()%360,0,0,1);
                                            			glBindTexture(GL_TEXTURE_2D,texture[2]);
                                            			glBegin(GL_TRIANGLE_STRIP);
                                            			    glTexCoord2d(1,1); glVertex3f(lsize,lsize,0);
                                            				glTexCoord2d(0,1); glVertex3f(-1*lsize,lsize,0);
                                            				glTexCoord2d(1,0); glVertex3f(lsize,-1*lsize,0);
                                            				glTexCoord2d(0,0); glVertex3f(-1*lsize,-1*lsize,0);
                                            			glEnd();
                                            			
                                            			glPopMatrix();
                                           				glDisable(GL_TEXTURE_2D);
                                				 	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                              				 	        glEnable(GL_BLEND);
                              				 	        glEnable(GL_DEPTH_TEST);
                                                                          				 	        
                                                        break;
                                                    case world::GAS:
                                                        glDisable(GL_DEPTH_TEST);
                                                        glBlendFunc(GL_SRC_ALPHA,GL_ONE);
                                                        glEnable(GL_TEXTURE_2D);
                                                        glBindTexture(GL_TEXTURE_2D,texture[0]);
                                            			glPushMatrix();
//                                            			GLfloat matrix[16];
                                            			glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
                                            			matrix[0] = 1.0f; matrix[1] = 0.0f; matrix[2] = 0.0f;
                                            			matrix[4] = 0.0f; matrix[5] = 1.0f; matrix[6] = 0.0f;
                                            			glLoadMatrixf(matrix);
                                            
                                            			glBegin(GL_TRIANGLE_STRIP);
                                            			    glTexCoord2d(1,1); glVertex3f(2,2,0);
                                            				glTexCoord2d(0,1); glVertex3f(-2,2,0);
                                            				glTexCoord2d(1,0); glVertex3f(2,-2,0);
                                            				glTexCoord2d(0,0); glVertex3f(-2,-2,0);
                                            			glEnd();
                                            			glPopMatrix();
                                           				glDisable(GL_TEXTURE_2D);
                                				 	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                              				 	        glEnable(GL_BLEND);
                              				 	        glEnable(GL_DEPTH_TEST);
                                                        break;
                                                }        
                                                break;
                                            case 1:
                                                glutSolidCube(1);
                                                break;
                                            case 2:
                                                glutSolidSphere(0.75,7,7);
                                                break;
                                        }    
                                    }    
                            glPopMatrix();
                            if (!pause){
                                w->step(a,b,c);
                            }    
                }    
            }
        }
    }
 
    glPushMatrix();
        glColor4f(1,1,1,1);
        glDisable(GL_DEPTH_TEST);
        render_text(w->types[selected_type].name,point_x+2,point_y+2,point_z+2);
        glEnable(GL_DEPTH_TEST);
    glPopMatrix();
    
    if (!pause){
    w->count_exp = (w->count_exp+1)%NUM_E;
    
        for(int i = 0; i < NUM_E; i++)
            exp[i]->step();
        delete exp[w->count_exp];
        exp[w->count_exp] = new fountain(fountain::EXPL);
    
        for(int i = 0; i < NUM_F; i++)
            f[i]->step();
        w->turn++;
    }    
    glutSwapBuffers();
}


static void key(unsigned char key, int x, int y){
    switch (key) {
        case 27 : 
        case 'q':
            for(int i = 0; i<NUM_F;i++)
                    delete f[i];
            exit(0);
            break;
        case 'l':
                add_sand();
            break;
        case 'p':
            cycle_type(-1,w->ptypes);
            break;
        case 's':
            graphics = (graphics + 1)%3;
            break;
        case 'a':
            polar_d -= 5;
            break;
        case 'z':
            polar_d += 5;
            break;
        case ' ':
            pause = !pause;
            break;
        case '=':
        case '+':
            pen_size++;
            if (pen_size > 7)
                        pen_size = 7;
            break;
        case '-':
            pen_size--;
            if (pen_size < 1)
                        pen_size = 1;
            break;
        case '1':
            selected_type = w->EMPTY;
            break;
        case '2':
            cycle_type(w->SOLID,w->ptypes);
            break;
        case '3':
            cycle_type(w->LIQUID,w->ptypes);
            break;
        case '4':
            cycle_type(w->GAS,w->ptypes);
            break;
        case '5':
            cycle_type(w->HOLLOW,w->ptypes);
            break;        
        case '6':
            cycle_type(w->SKYHOOK,w->ptypes);
            break; 
        case '7':
            cycle_type(w->GENERATOR,w->ptypes);
            break;        
    }
    glutPostRedisplay();
}


static void speckey(int key, int x, int y){
    switch (key) {
       case GLUT_KEY_F1:
       polar_x = 90;
       polar_y = 0;
       break;
       
       case GLUT_KEY_F2:
       polar_x = 20;
       polar_y = 0;
       break;
       
       case GLUT_KEY_F3:
       polar_y -= 90;
       break;
       
       case GLUT_KEY_F4:
       polar_y += 90;
       break;
        
       case GLUT_KEY_LEFT:
           point_x--;
       break;

       case GLUT_KEY_UP:
           point_z--;
       break;
       
       case GLUT_KEY_RIGHT:
           point_x++;
       break;
       
       case GLUT_KEY_DOWN:
           point_z++;
       break;
       
       case GLUT_KEY_PAGE_UP:
           point_y += pen_size;
       break;

       case GLUT_KEY_PAGE_DOWN:
           point_y -= pen_size;
       break;
       
       case GLUT_KEY_F5:  
       break;
       
       case GLUT_KEY_F10:
           delete w;
           w = new world(a,b,c);
       break;
    }
    if (point_x < 0) point_x = 0;
    if (point_y < 0) point_y = 0;
    if (point_z < 0) point_z = 0;
    if (point_x > w->x-1) point_x = w->x-1;
    if (point_y > w->y-1) point_y = w->y-1;
    if (point_z > w->z-1) point_z = w->z-1;  
    glutPostRedisplay();
}   

static void pmouse(int x, int y){
    int difx, dify;
    difx = mouse_x - x;
    dify = mouse_y - y;
    mouse_x = x;
    mouse_y = y;
    if (glutGetModifiers() == GLUT_ACTIVE_CTRL) return;
    if (camera_on){
        polar_x += (int)dify*0.75;
        polar_y -= (int)difx*0.75;
    } else if (glutGetModifiers() == GLUT_ACTIVE_SHIFT){
        point_y += (int)(dify*0.5);
        
        if (point_y < 0) point_y = 0;
        if (point_y > w->y-1) point_y = w->y-1;    
    } else {
        /*        
        point_x -= difx;  
        point_z -= dify;
          
        if (point_x >= w->x-1) point_x = w->x-2;
        if (point_x < 0) point_x=0;
        if (point_z >= w->z) point_z = w->z-2;
        if (point_z < 0) point_z=0;
        */


        double objx, objy, objz;
        double px, py, pz;
        
        double modelview[16], projection[16];
        int viewport[4];
        float z;
        	//get the projection matrix		
        glGetDoublev( GL_PROJECTION_MATRIX, projection );
        	//get the modelview matrix		
        glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
        	//get the viewport		
        glGetIntegerv( GL_VIEWPORT, viewport );
        	
        //Read the window z co-ordinate 
        //(the z value on that point in unit cube)		
        glReadPixels( x, viewport[3]-y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &z );
        
        //Unproject the window co-ordinates to 
        //find the world co-ordinates.
        gluUnProject( x, viewport[3]-y, z, modelview, projection, viewport, &objx, &objy, &objz );

        point_x = (int)objx;
        //point_y = (int)objy+1;
        point_z = (int)objz;
        
        if (point_x < 0) point_x = 0;
        //if (point_y < 0) point_y = 0;
        if (point_z < 0) point_z = 0;
        if (point_x > w->x-1) point_x = w->x-1;
        //if (point_y > w->y-1) point_y = w->y-1;
        if (point_z > w->z-1) point_z = w->z-1;  
    }    
    
}

static void mouse(int button, int state, int x, int y){
   switch(button) {
   case GLUT_LEFT_BUTTON:
      if (state == GLUT_DOWN){
         sand_on = true; 
      } else {
         sand_on = false; 
      }        
      break;
   case GLUT_RIGHT_BUTTON:
      if (state == GLUT_DOWN){
         camera_on = true; 
      } else {
         camera_on = false;
      }  
      break;
   case GLUT_WHEEL_UP:
       if ( state == GLUT_UP){
            polar_d -= 5;
       }
       break;
   case GLUT_WHEEL_DOWN:
       if ( state == GLUT_UP){
           polar_d += 5;

       }      
       break;

   }

    pmouse(x,y);    
    
}

static void idle(void){
    glutPostRedisplay();
}

/* Program entry point */

int main(int argc, char *argv[]) {
    if (argc == 4){
      a = atoi(argv[1]);  
      b = atoi(argv[2]);
      c = atoi(argv[3]);
      if (a < 1 || a > 1000) a = 100;
      if (b < 1 || b > 1000) b = 50;
      if (c < 1 || c > 1000) c = 100;
    }
    
    windx = 640;
    windy = 480;
    glutInit(&argc, argv);
    glutInitWindowSize(windx,windy);
    glutInitWindowPosition(10,10);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    char title[64];
    sprintf(title,"Sand3d %g - David Pascoe-Deslauriers, 2007",VERSION);
    glutCreateWindow(title);

    glutReshapeFunc(resize);
    glutDisplayFunc(display);
    glutKeyboardFunc(key);
    glutSpecialUpFunc(speckey);
    glutMouseFunc(mouse);
    glutMotionFunc(pmouse);
    glutPassiveMotionFunc(pmouse);

    glutIdleFunc(idle);

    glClearColor(0,0,0,1);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glEnable(GL_LIGHT0);
    glEnable(GL_NORMALIZE);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);
    
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    glAlphaFunc(GL_GREATER,0.1f);
    glEnable(GL_ALPHA_TEST);
    
    w = new world(a,b,c);

    loadGLTextures("particle1.bmp",&texture[0]);
    loadGLTextures("particle2.bmp",&texture[1]);
    loadGLTextures("particle3.bmp",&texture[2]);
    loadGLTextures("bg.bmp",&texture[3]);
 
    for(int i = 0; i < NUM_F; i++)
        f[i] = new fountain(fountain::FIRE);
    
    w->count_exp = 0;
    for(int i = 0; i < NUM_E; i++)
        exp[i] = new fountain(fountain::EXPL);       
            
    sand_on = false;
    camera_on = false;
    pause = false;
    pen_size = 3;
    graphics = 0;
    polar_x = 20;
    polar_y = 0;
    polar_d = 100;
    point_x = 0;
    point_y = w->y/2;
    point_z = 0;
    
    fps =0;
    frames=0;
    timeEl = 0;
    particleCount = 0;
    selected_type = w->SAND;
     
    glutMainLoop();

    return EXIT_SUCCESS;
}

