/*
Sand3d
Copyright (C) 2007 - David Pascoe-Deslauriers

  This file is part of Sand3d.

    Sand3d is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Sand3d is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sand3d; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "fountain.h"
#include "octtree.h"

#define VERSION 0.44

/* Structure for storing 3d vectors */
struct vector {
    double x;
    double y;
    double z;
} typedef vector;


/* Structure for storing Particles */
struct particle {
    int uniqueid;
    char name[24];
    int fireid;
    int turn;
    double mass;
    int type;
    vector colour;
    int combustion_counter;
    double alpha;
    bool puts_out_fire;
    int fire;
    int phase;
    bool force;
    int secondary_counter;
    int yrot;
    bool explodes;
    bool destroy;
} typedef particle;    


class world {
    public:
    int x;    /* Depth */
    int y;     /* Height */
    int z;    /* Width */
    
    /* Define particle types; These will not be defined here once particle type our setup to load from external files */

    static const int EMPTY = 0;
    static const int SAND = 1;
    static const int OIL = 2;
    static const int WATER = 3;
    static const int METHANE = 4;
    static const int FIRE = 5;
    static const int GLASS = 6;
    static const int WALL = 7;
    static const int H2S = 8;
    static const int CO2 = 9;
    static const int N2O = 10;
    static const int TORCH = 11;
    static const int SPOUT = 12;
    static const int WELL = 13;
    static const int COW = 14;    
    static const int GUNPOWDER = 15;
    static const int C4 = 16;
    static const int NAPALM = 17;
    

    /* Define state of matter for the particles */
    
    static const int SOLID = 1;
    static const int LIQUID = 2;
    static const int GAS = 3;
    static const int SKYHOOK = 4;
    static const int HOLLOW = 5;
    static const int GENERATOR = 6;
    
    
    /* Default particle types */
    static const int ptypes = 18;
    particle* types;
    int turn;
    /* Storage for particles in the game world*/
    particle ***w;
    
    data_strc *active;
    
    int count_exp;
    
        world(int,int,int);
        ~world();
        
        int add_particle(int,int,int,int);
        
        int step(int,int,int);
        void motion(int,int,int);
        void gas_motion(int, int, int);
        void spread_fire(int,int,int);
        void generate(int,int,int);

        particle get_particle(int,int,int);
        
        void swap(int,int,int,int,int,int);


        /* Functions for loading user made particles from files; Not Finished */
        void SaveGame();
        
        particle getParticleFrom(char *file, int id);
           
};    
